/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <strings.h>

#include <gandalf/image.h>
#include <gandalf/image/image_defs.h>
#include <gandalf/image/io/image_io.h>
#include <gandalf/image/image_display.h>

#include "vis_timing.h"
#define NEIGHBOUR 1
#define BILINEAR1 2
#define BILINEAR2 3

void undistortImageBilinear2(Gan_Image *img, Gan_Image * undistImage, int **lutxFloor, int ** lutyFloor,
			     float **lutxDiff, float ** lutyDiff, int width, int height)
{  
  int i,j;

  int *uip_f,*ujp_f;
  float *dip,*djp;
  unsigned char **row_ptr=(unsigned char**)undistImage->row_data_ptr;
  unsigned char *row;
  int ui_f,uj_f,ui_c,uj_c;
  float di,dj;
  unsigned char value,val_ff,val_fc,val_cf,val_cc;
  
  for(j=0;j<height;j++) {
    ujp_f = lutxFloor[j];
    uip_f = lutyFloor[j];
    djp = lutxDiff[j];
    dip = lutyDiff[j];
    row=row_ptr[j];
    
    for(i=0;i<width;i++) {
      uj_f = ujp_f[i];
      ui_f = uip_f[i];

      uj_c = uj_f+1;
      ui_c = ui_f+1;
      
      dj = djp[i];
      di = dip[i];

      /*if (dj!=0)
	printf("%1.30 f ",dj);
      */
      if ((ui_c<width)&&(ui_f>=0)&&
	  (uj_c<height)&&(uj_f>=0)) {
	val_ff = gan_image_get_pix_gl_uc(img, uj_f, ui_f);
	val_fc = gan_image_get_pix_gl_uc(img, uj_f, ui_c);
	val_cf = gan_image_get_pix_gl_uc(img, uj_c, ui_f);
	val_cc = gan_image_get_pix_gl_uc(img, uj_c, ui_c);
	
	row[i]= val_ff+
	  dj*(val_cf-val_ff)+
	  di*(val_fc-val_ff)+
	  dj*di*(val_ff+val_cc-val_cf-val_fc);
      }
    }
  }
} 

void undistortImageBilinear1(Gan_Image *img, Gan_Image * undistImage, float **lutx, float **luty,int width, int height)
{  
  int i,j;

  float *uip_f,*ujp_f;
  float *dip,*djp;
  unsigned char **row_ptr=(unsigned char**)undistImage->row_data_ptr;
  unsigned char *row;
  int ui_f,uj_f,ui_c,uj_c;
  float di,dj;
  unsigned char value,val_ff,val_fc,val_cf,val_cc;
  
  for(j=0;j<height;j++) {
    ujp_f = lutx[j];
    uip_f = luty[j];
    row=row_ptr[j];
    
    for(i=0;i<width;i++) {
      uj_f = (int)floor(ujp_f[i]);
      ui_f = (int)floor(uip_f[i]);

      uj_c = uj_f+1;
      ui_c = ui_f+1;
      
      dj = ujp_f[i]-uj_f;
      di = uip_f[i]-ui_f;

      /*if (dj!=0)
	printf("%1.30 f ",dj);
      */
      if ((ui_c<width)&&(ui_f>=0)&&
	  (uj_c<height)&&(uj_f>=0)) {
	val_ff = gan_image_get_pix_gl_uc(img, uj_f, ui_f);
	val_fc = gan_image_get_pix_gl_uc(img, uj_f, ui_c);
	val_cf = gan_image_get_pix_gl_uc(img, uj_c, ui_f);
	val_cc = gan_image_get_pix_gl_uc(img, uj_c, ui_c);
	
	row[i]= val_ff+
	  dj*(val_cf-val_ff)+
	  di*(val_fc-val_ff)+
	  dj*di*(val_ff+val_cc-val_cf-val_fc);
      }
    }
  }
} 


void undistortImageNearest(Gan_Image *img, Gan_Image * undistImage, int **lutx, int ** luty, int width, int height)
{
  int i,j;
  int *uip,*ujp;
  unsigned char **row_ptr=(unsigned char**)undistImage->row_data_ptr;
  unsigned char *row;

  gan_image_get_pix_gl_uc(img, 4, 473);

  for(j=0;j<height;j++) {
    uip = lutx[j];
    ujp = luty[j];
    row=row_ptr[j];
    
    for(i=0;i<width;i++) {
      // The values are known to be in the image
      //if ((ui<width)&&(ui>=0)&&
      //    (uj<height)&&(uj>=0)) 
      //printf("%i,%i=%i,%i ", i,j,uip[i], ujp[i]);
      row[i]=gan_image_get_pix_gl_uc(img, uip[i], ujp[i]);
    }   
  }
}

float **loadLutFloat(char *fileName, int *width, int *height) {
  FILE * fid=fopen (fileName, "r");

  if(fid==NULL) {
    fprintf(stderr,"Error opening file '%s'\n",fileName);
    exit(-1);
  }
  
  fread(width, sizeof(int), 1, fid);
  fread(height, sizeof(int), 1, fid);
  
  float **lut = malloc(*height*sizeof(float*));
  int i;

  for(i=0;i<*height;i++) {
    lut[i] = malloc(sizeof(float)*(*width));
    fread(lut[i],sizeof(float),*width,fid);
  }

  fclose(fid);

  return lut;
}

int **loadLutInt(char *fileName, int *width, int *height) {
  FILE * fid=fopen (fileName, "r");

  if(fid==NULL) {
    fprintf(stderr,"Error opening file '%s'\n",fileName);
    exit(-1);
  }

  fread(width, sizeof(int), 1, fid);
  fread(height, sizeof(int), 1, fid);
  
  int **lut = malloc(*height*sizeof(int*));
  int i;

  for(i=0;i<*height;i++) {
    lut[i] = malloc(sizeof(int)*(*width));
    fread(lut[i],sizeof(int),*width,fid);
  }

  fclose(fid);

  return lut;
}

void checkLutSizes(int width1,int width2,int height1,int height2, const char *msg) {
  if ((width1!=width2)||
      (height1!=height2)) {
    printf(msg);
    exit(-1);
  }
}

void checkImage(Gan_Image *img,char *inputImage) {
  if(img==NULL) {
    printf("Problem loading image '%s'.\n",inputImage);
    exit(-1);
  }
}

void printUsageAndExit() {
  printf("USAGE : undistort -n  lutxFileRound lutyFileRound inputImage outputImage\n");
  printf("   or : undistort -b1 lutx luty inputImage outputImage\n");
  printf("   or : undistort -b2 lutxFileFloor lutyFileFloor lutxFileDiff lutyFileDiff inputImage outputImage\n");
  exit(-1);
}

Gan_ImageFileFormat fileFormatFromName(char *fileName) {
  char * pointerP = rindex(fileName, 'p');

  if(pointerP!=NULL) {
    if(strlen(pointerP)<4) {
      if(strcmp(pointerP,"ppm")==0) {
	//printf("Found PPM file type.\n");
	return GAN_PPM_FORMAT;
      }
      if(strcmp(pointerP,"pgm")==0) {
	//printf("Found PGM file type.\n");
	return GAN_PGM_FORMAT;
      }
      if(strcmp(pointerP,"png")==0) {
	//printf("Found PNG file type.\n");
	return GAN_PNG_FORMAT;
      }
      if(strcmp(pointerP,"peg")==0) {
	//printf("Found JPEG file type.\n");
	return GAN_JPEG_FORMAT;
      }
      if(strcmp(pointerP,"pg")==0) {
	//printf("Found JPEG file type.\n");
	return GAN_JPEG_FORMAT;
      }
      printf("Unknown file type based on name, assuming PGM.\n");
      return GAN_PGM_FORMAT;
    }
  }

  pointerP = rindex(fileName, 't');

  if(pointerP!=NULL) {
    if(strlen(pointerP)<4) {
      if(strcmp(pointerP,"tif")==0) {
	//printf("Found TIFF file type.\n");
	return GAN_TIFF_FORMAT;
      }
    }
  }
  printf("Unknown file type based on name, assuming PGM.\n");
  return GAN_PGM_FORMAT;
}

int main( int argc, char *argv[] )
{  
  if (argc<6) {
    printUsageAndExit();
  }

  struct timeval tv;

  int bilinear;

  if (strcmp(argv[1],"-n")==0) {
    bilinear = NEIGHBOUR;
  } else {
    if (strcmp(argv[1],"-b1")==0) {
      bilinear = BILINEAR1;
    } else {
	bilinear = BILINEAR2;
    }
  }

  char * inputImage;
  char * outputImage;
  int lutx_width, lutx_height, luty_width, luty_height;
  Gan_Image *img;
  int width, height;
  double t1;
  Gan_Image * undistImage;
  Gan_ImageFileFormat inputFileFormat, outputFileFormat;

  switch(bilinear) {
  case NEIGHBOUR:
    if(argc!=6)
      printUsageAndExit();

    char * lutxFileRound = argv[2];
    char * lutyFileRound = argv[3];
    inputImage = argv[4];
    outputImage = argv[5];

    inputFileFormat = fileFormatFromName(inputImage);
    outputFileFormat = fileFormatFromName(inputImage);

    //printf("LutxFileRound : %s\n", lutxFileRound);
    //printf("LutyFileRound : %s\n", lutyFileRound);
    //printf("InputImage : %s\n", inputImage);
    //printf("OutputImage : %s\n", outputImage);

    t1=REALTIME;

    int **lutxRound = loadLutInt(lutxFileRound,&lutx_width,&lutx_height);
    int **lutyRound = loadLutInt(lutyFileRound,&luty_width,&luty_height);

    checkLutSizes(lutx_width,luty_width,lutx_height,luty_height,"The LUT tables don't have the same size.\n");

    //Load image
    img = gan_image_read ( inputImage, inputFileFormat, NULL );
    checkImage(img,inputImage);
    
    width = img->width;
    height = img->height;
    
    checkLutSizes(width,lutx_width,height,lutx_height,"LUT tables don't have the same size as the image.\n");
    
    undistImage = gan_image_form_gen(NULL, img->format, img->type, height,
				     width, img->stride, GAN_TRUE, NULL, 0,NULL,0);

    t1=REALTIME;
    undistortImageNearest(img,undistImage,lutxRound,lutyRound,width,height);

    //printf("Time taken to undist : %2.5fs\n",REALTIME-t1);

    gan_image_write ( outputImage, outputFileFormat, undistImage);

    break;
  case BILINEAR1:
    if(argc!=6)
      printUsageAndExit();

    char * lutxFile = argv[2];
    char * lutyFile = argv[3];
    inputImage = argv[4];
    outputImage = argv[5];

    inputFileFormat = fileFormatFromName(inputImage);
    outputFileFormat = fileFormatFromName(inputImage);

    //printf("LutxFile : %s\n", lutxFile);
    //printf("LutyFile : %s\n", lutyFile);
    //printf("InputImage : %s\n", inputImage);
    //printf("OutputImage : %s\n", outputImage);
    
    t1=REALTIME;

    float **lutx = loadLutFloat(lutxFile,&lutx_width,&lutx_height);
    float **luty = loadLutFloat(lutyFile,&luty_width,&luty_height);

    checkLutSizes(lutx_width,luty_width,lutx_height,luty_height,"The LUT tables don't have the same size.\n");
 
    //Load image
    Gan_Image *img = gan_image_read ( inputImage, inputFileFormat, NULL );
    checkImage(img,inputImage);
    
    width = img->width;
    height = img->height;
    
    checkLutSizes(width,lutx_width,height,lutx_height,"LUT tables don't have the same size as the image.\n");

    undistImage = gan_image_form_gen(NULL, img->format, img->type, height,
				     width, img->stride, GAN_TRUE, NULL, 0,NULL,0);


    t1=REALTIME;
    undistortImageBilinear1(img,undistImage,lutx,luty,width,height);
  
    //printf("Time taken to undist : %2.5fs\n",REALTIME-t1);

    gan_image_write ( outputImage, outputFileFormat, undistImage);

    break;
  case BILINEAR2:
    if(argc!=8)
      printUsageAndExit();
    
    char * lutxFileFloor = argv[2];
    char * lutyFileFloor = argv[3];
    char * lutxFileDiff = argv[4];
    char * lutyFileDiff = argv[5];
    inputImage = argv[6];
    outputImage = argv[7];

    inputFileFormat = fileFormatFromName(inputImage);
    outputFileFormat = fileFormatFromName(inputImage);

    //printf("LutxFileFloor : %s\n", lutxFileFloor);
    //printf("LutyFileFloor : %s\n", lutyFileFloor);
    //printf("LutxFileDiff : %s\n", lutxFileDiff);
    //printf("LutyFileDiff : %s\n", lutyFileDiff);
    //printf("InputImage : %s\n", inputImage);
    //printf("OutputImage : %s\n", outputImage);

    t1=REALTIME;

    int **lutxFloor = loadLutInt(lutxFileFloor,&lutx_width,&lutx_height);
    int **lutyFloor = loadLutInt(lutyFileFloor,&luty_width,&luty_height);
 
    checkLutSizes(lutx_width,luty_width,lutx_height,luty_height,"The LUT tables don't have the same size.\n");

    float **lutxDiff = loadLutFloat(lutxFileDiff,&lutx_width,&lutx_height);
    float **lutyDiff = loadLutFloat(lutyFileDiff,&luty_width,&luty_height);

    checkLutSizes(lutx_width,luty_width,lutx_height,luty_height,"The LUT tables don't have the same size.\n");    

    //printf("Time taken for loading luts : %2.5fs\n",REALTIME-t1);

    //Load image
    img = gan_image_read ( inputImage, inputFileFormat, NULL );
    checkImage(img,inputImage);

    width = img->width;
    height = img->height;

    checkLutSizes(width,lutx_width,height,lutx_height,"LUT tables don't have the same size as the image.\n");

    /*printf("Format : %i\n",img->format);
      printf("GREY_LEVEL : %i\n",GAN_GREY_LEVEL_IMAGE);
      printf("Type : %i\n",img->type);
      printf("UNSIGNED CHAR : %i\n",GAN_UNSIGNED CHAR);
    */
    undistImage = gan_image_form_gen(NULL, img->format, img->type, height,
				     width, img->stride, GAN_TRUE, NULL, 0,NULL,0);

    t1=REALTIME;

    undistortImageBilinear2(img,undistImage,lutxFloor,lutyFloor,lutxDiff,lutyDiff,width,height);
  
    //printf("Time taken to undist : %2.5fs\n",REALTIME-t1);
    
    gan_image_write ( outputImage, outputFileFormat, undistImage);
    
    break;
  }

  return 0;
}
